"""
Functions that process data for plotting
"""

import math
from typing import List, Union, Optional, Tuple, Dict

from obspy import UTCDateTime

from sohstationviewer.conf import constants as const
from sohstationviewer.view.util.enums import LogType

MAX_INT = 1E100
MAX_FLOAT = 1.0E100

# TODO: put this in DB
mass_pos_volt_ranges = {"regular": [0.5, 2.0, 4.0, 7.0],
                        "trillium": [0.5, 1.8, 2.4, 3.5]}
mass_pos_color_pallets = {"B": ["C", "G", "Y", "R", "M"],
                          "W": ["B", "B", "B", "B", "B"]}


def get_masspos_value_colors(
        range_opt: str, chan_id: str, c_mode: str,
        processing_log: List[Tuple[str, LogType]],
        ret_type: str = 'str'
) -> Optional[Union[str, List[Tuple[float, str]]]]:
    """
    Create a map between value and color based on given rangeOpt and c_mode to
    display mass position plots.
    :param range_opt: massPosVoltRangeOpt got from Options Menu - MP coloring
        in Main Window to define different values for mass position.
        (regular/trillium)
    :param chan_id: ID of the channel
    :param c_mode: color mode (B/W)
    :param processing_log: list of processing info and type
    :param ret_type: request return type
    :return: [(value, color), (value, color) ...]
        if retType is 'str', return "value:color|value:color"
    """

    if range_opt.lower() not in mass_pos_volt_ranges.keys():
        processing_log.append(
            (
                f"{chan_id}: The current selected Mass Position color range is"
                f" '{range_opt}' isn't allowable. The accept ranges are: "
                f"{', '.join(mass_pos_volt_ranges.keys())}",
                LogType.ERROR
            )
        )
        return
    mass_pos_volt_range = mass_pos_volt_ranges[range_opt]
    mass_pos_color_pallet = mass_pos_color_pallets[c_mode]
    value_colors = []
    for i in range(len(mass_pos_volt_range)):
        if ret_type == 'str':
            value_colors.append(
                "%s:%s" % (mass_pos_volt_range[i], mass_pos_color_pallet[i]))
        else:
            value_colors.append(
                (mass_pos_volt_range[i], mass_pos_color_pallet[i])
            )
        if i == len(mass_pos_volt_range) - 1:
            if ret_type == 'str':
                value_colors.append(
                    "%s:+%s" % (
                        mass_pos_volt_range[i], mass_pos_color_pallet[i + 1]))
            else:
                value_colors.append(
                    (mass_pos_volt_range[i], mass_pos_color_pallet[i + 1]))
    if ret_type == 'str':
        return '|'.join(value_colors)
    return value_colors


def format_time(time: Union[UTCDateTime, float], date_mode: str,
                time_mode: Optional[str] = None) -> str:
    """
    Format time according to date_mode and time_mode
    :param time: time to be format, can be UTCDateTime or epoch time
    :param date_mode: the format of date
    :param time_mode: the format of time
    :return: the formatted time string
    """
    if isinstance(time, UTCDateTime):
        t = time
    else:
        t = UTCDateTime(time)

    # https://docs.python.org/3/library/datetime.html#
    # strftime-and-strptime-format-codes
    format = ''
    if date_mode == 'YYYY-MM-DD':
        format = '%Y-%m-%d'
    elif date_mode == 'YYYYMMMDD':
        format = '%Y%b%d'
    elif date_mode == 'YYYY:DOY':
        format = '%Y:%j'
    if time_mode == 'HH:MM:SS':
        format += " %H:%M:%S"

    ret = t.strftime(format).upper()
    return ret


def get_title(key: Union[str, Tuple[str, int]], min_time: float,
              max_time: float, date_mode: str) -> str:
    """
    Create title for the plot.
    :param key: str or (str, str) sta_id for mseed, (unit_id, exp_no) for rt130
    :param min_time: start time of the plot
    :param max_time: end time of the plot
    :param date_mode: format of date
    :return: title for the plot
    """
    diff = max_time - min_time
    hours = diff / 3600
    return ("%s  %s  to  %s  (%s)" %
            (key,
             format_time(min_time, date_mode, "HH:MM:SS"),
             format_time(max_time, date_mode, "HH:MM:SS"),
             round(hours, 2))
            )


def get_gaps(gaps: List[List[float]], gap_min: float
             ) -> List[List[float]]:
    """
    :param gaps: list of gaps
    :param gap_min: minimum of gaps count in minutes
    :return: list of gaps of which gaps smaller than gapMin have been removed
    """
    gap_min_sec = gap_min * 60
    return [g for g in gaps if abs(g[1] - g[0]) >= gap_min_sec]


def get_time_ticks(earliest: float, latest: float, date_format: str,
                   label_total: int
                   ) -> Tuple[List[float], List[float], List[str]]:
    """
    split time range to use for tick labels
    Ex: getTimeTicks(1595542860.0, 1595607663.91, 'YYYY-MM-DD', 3)
    :param earliest: earliest epoch time
    :param latest: latest epoch time
    :param date_format: (YYYY:DOY, YYYY-MM-DD or YYYYMMMDD)
        (selected in Menu Options - Date Format)
    :param label_total: number of time label to be displayed,
        others will show as ticks oly
    :return:
        times: list of times to show ticks
        major_times: list of times for displayed labels
        major_time_labels: list of labels displayed
    """
    time_range = latest - earliest
    if time_range >= 2592000.0:
        mode = "DD"
        # Time of the nearest midnight before the earliest time.
        time = (earliest // 86400.0) * 86400
        interval = 864000.0
    elif time_range >= 864000.0:
        mode = "D"
        time = (earliest // 86400.0) * 86400
        interval = 86400.0
    elif time_range >= 3600.0:
        mode = "H"
        # Nearest hour.
        time = (earliest // 3600.0) * 3600
        interval = 3600.0
    elif time_range >= 60.0:
        mode = "M"
        # Nearest minute.
        time = (earliest // 60.0) * 60
        interval = 60.0
    else:
        mode = "S"
        time = (earliest // 1)
        interval = 1.0
    times = []
    time_labels = []
    time += interval
    while time <= latest:
        times.append(time)
        time_label = format_time(time, date_format, 'HH:MM:SS')
        if mode == "DD" or mode == "D":
            time_label = time_label[:-9]
        elif mode == "H":
            time_label = time_label[:-3]
        elif mode == "M" or mode == "S":
            time_label = time_label
        time_labels.append(time_label)
        time += interval

    ln = len(time_labels)
    d = math.ceil(len(time_labels) / label_total)
    major_times = [times[i] for i in range(ln) if i % d == 0]
    major_time_labels = [time_labels[i] for i in range(ln) if i % d == 0]
    return times, major_times, major_time_labels


def get_day_ticks() -> Tuple[List[int], List[int], List[str]]:
    """
    Get information for displaying time on plotting widget.
    :return:
        times: list of indexes of every hour in each_day_5_min_list
        major_times: list of indexes of every 4 hours in each_day_5_min_list
        major_time_labels: 2 digit numbers of every 4 hours in a day
    """

    times = list(range(const.NO_5M_1H, const.NO_5M_DAY, const.NO_5M_1H))
    major_times = list(range(4 * const.NO_5M_1H,
                             const.NO_5M_DAY,
                             4 * const.NO_5M_1H))
    major_time_labels = ["%02d" % int(t / const.NO_5M_1H) for t in major_times]
    return times, major_times, major_time_labels


def get_unit_bitweight(chan_db_info: Dict, bitweight_opt: str) -> str:
    """
    Get the format to display value including fixed point decimal and unit
    based on the information from the database.
    :param chan_db_info: channel's info got from database
    :param bitweight_opt: option for bitweight (none, low, high)
        (Menu Options - Q330 Gain)
    :return unit_bitweight: format for displayed value on the left of each
        plot with fixed point decimal and unit
    """

    plot_type = chan_db_info['plotType']
    # Not all channels use/have this field.
    unit = chan_db_info['unit']
    try:
        fix_point = chan_db_info['fixPoint']
    except Exception:
        fix_point = 0
    unit_bitweight = ''

    if plot_type in ['linesDots', 'linesSRate', 'linesMasspos']:

        if fix_point == 0:
            unit_bitweight = "{}%s" % unit
        else:
            unit_bitweight = "{:.%sf}%s" % (fix_point, unit)
    if chan_db_info['channel'] == 'SEISMIC':
        if fix_point != 0:
            unit_bitweight = "{:.%sf}%s" % (fix_point, unit)
        else:
            if bitweight_opt in ["low", "high"]:
                unit_bitweight = "{}V"
            else:
                unit_bitweight = "{}%s" % unit
    return unit_bitweight
