from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Dict, Union, List

from unittest import TestCase
from unittest.mock import patch

from obspy.core import UTCDateTime
import numpy as np

from sohstationviewer.conf import constants as const
from sohstationviewer.model.handling_data import (
    trim_downsample_chan_with_spr_less_or_equal_1,
    trim_downsample_wf_chan,
    trim_waveform_data,
    downsample_waveform_data,
)
from sohstationviewer.model.downsampler import downsample, chunk_minmax

ORIGINAL_CHAN_SIZE_LIMIT = const.CHAN_SIZE_LIMIT
ORIGINAL_RECAL_SIZE_LIMIT = const.RECAL_SIZE_LIMIT
ZERO_EPOCH_TIME = UTCDateTime(1970, 1, 1, 0, 0, 0).timestamp
TraceInfo = Dict[str, Union[float, str]]
ChannelData = Dict[str,
                   Union[float, List[np.ndarray], List[TraceInfo], Dict,
                         np.ndarray]]


class TestTrimWfData(TestCase):
    def setUp(self) -> None:
        self.channel_data = {}
        self.traces_info = []
        self.channel_data['tracesInfo'] = self.traces_info

        for i in range(100):
            trace_size = 100
            start_time = i * trace_size
            trace = {}
            trace['startTmEpoch'] = start_time
            trace['endTmEpoch'] = start_time + trace_size - 1
            self.traces_info.append(trace)
        self.start_time = 2500
        self.end_time = 7500

    def test_data_is_trimmed_neither_start_nor_end_time_is_trace_start_or_end_time(self):  # noqa: E501
        self.start_time = 2444
        self.end_time = 7444
        trimmed_traces_list = trim_waveform_data(
            self.channel_data, self.start_time, self.end_time
        )

        self.assertTrue(
            trimmed_traces_list[0]['startTmEpoch'] <= self.start_time)
        self.assertTrue(
            trimmed_traces_list[0]['endTmEpoch'] > self.start_time
        )
        trimmed_traces_list.pop(0)
        trimmed_traces_list.pop()
        is_left_trimmed = all(trace['startTmEpoch'] > self.start_time
                              for trace in trimmed_traces_list)
        is_right_trimmed = all(trace['endTmEpoch'] <= self.end_time
                               for trace in trimmed_traces_list)
        self.assertTrue(is_left_trimmed and is_right_trimmed)

    def test_data_out_of_range(self):
        with self.subTest('test_start_time_later_than_data_end_time'):
            self.start_time = 12500
            self.end_time = 17500
            self.assertFalse(
                trim_downsample_wf_chan(self.channel_data, self.start_time,
                                        self.end_time, True)
            )
        with self.subTest('test_end_time_earlier_than_data_start_time'):
            self.start_time = -7500
            self.end_time = -2500
            self.assertFalse(
                trim_downsample_wf_chan(self.channel_data, self.start_time,
                                        self.end_time, True)
            )

    def test_no_data(self):
        self.channel_data['tracesInfo'] = []
        with self.assertRaises(IndexError):
            trim_waveform_data(
                self.channel_data, self.start_time, self.end_time
            )

    def test_end_time_earlier_than_start_time(self):
        self.start_time, self.end_time = self.end_time, self.start_time
        trimmed_traces_list = trim_waveform_data(
            self.channel_data, self.start_time, self.end_time
        )
        self.assertListEqual(trimmed_traces_list, [])

    def test_data_does_not_need_to_be_trimmed(self):
        with self.subTest('test_start_time_earlier_than_trace_earliest_time'):
            self.start_time = -2500
            self.end_time = 7500
            trimmed_traces_list = trim_waveform_data(
                self.channel_data, self.start_time, self.end_time
            )
            self.assertEqual(len(trimmed_traces_list), 76)
        with self.subTest('test_end_time_later_than_trace_latest_time'):
            self.start_time = 2500
            self.end_time = 12500
            trimmed_traces_list = trim_waveform_data(
                self.channel_data, self.start_time, self.end_time
            )
            self.assertEqual(len(trimmed_traces_list), 75)
        with self.subTest('test_data_contained_in_time_range'):
            self.start_time = self.traces_info[0]['startTmEpoch']
            self.end_time = self.traces_info[-1]['endTmEpoch']
            trimmed_traces_list = trim_waveform_data(
                self.channel_data, self.start_time, self.end_time
            )
            self.assertEqual(len(trimmed_traces_list), len(self.traces_info))


class TestDownsampleWaveformData(TestCase):
    def no_file_memmap(self, file_path: Path, **kwargs):
        # Data will look the same as times. This has two benefits:
        # - It is a lot easier to inspect what data remains after trimming
        # and downsampling, seeing as the remaining data would be the same
        # as the remaining times.
        # - It is a lot easier to reproducibly create a test data set.
        array_size = 100
        file_idx = int(file_path.name.split('-')[-1])
        start = file_idx * array_size
        end = start + array_size
        return np.arange(start, end)

    def setUp(self) -> None:
        memmap_patcher = patch.object(np, 'memmap',
                                      side_effect=self.no_file_memmap)
        self.addCleanup(memmap_patcher.stop)
        memmap_patcher.start()

        self.channel_data = {}
        self.traces_info = []
        self.channel_data['tracesInfo'] = self.traces_info
        self.data_folder = TemporaryDirectory()
        for i in range(100):
            trace_size = 100
            start_time = i * trace_size
            trace = {}
            trace['startTmEpoch'] = start_time
            trace['endTmEpoch'] = start_time + trace_size - 1
            trace['size'] = trace_size

            times_file_name = Path(self.data_folder.name) / f'times-{i}'
            trace['times_f'] = times_file_name

            data_file_name = Path(self.data_folder.name) / f'data-{i}'
            trace['data_f'] = data_file_name

            self.traces_info.append(trace)
        self.start_time = 2550
        self.end_time = 7550
        self.trimmed_traces_list = trim_waveform_data(
            self.channel_data, self.start_time, self.end_time
        )

    @patch('sohstationviewer.model.handling_data.downsample', wraps=downsample)
    def test_data_is_downsampled(self, mock_downsample):
        const.CHAN_SIZE_LIMIT = 1000
        downsample_waveform_data(self.trimmed_traces_list,
                                 self.start_time, self.end_time)
        self.assertTrue(mock_downsample.called)
        const.CHAN_SIZE_LIMIT = ORIGINAL_CHAN_SIZE_LIMIT

    def test_all_traces_handled(self):
        downsampled_times, downsampled_data = downsample_waveform_data(
            self.trimmed_traces_list,
            self.start_time, self.end_time
        )
        self.assertEqual(len(downsampled_times), 51)
        self.assertEqual(len(downsampled_data), 51)

    def test_downsampling_not_needed(self):
        downsampled_times, downsampled_data = downsample_waveform_data(
            self.trimmed_traces_list,
            self.start_time, self.end_time
        )
        with self.subTest('test_data_points_outside_time_range_removed'):
            self.assertEqual(downsampled_times.pop(0).size, 50)
            self.assertEqual(downsampled_times.pop(-1).size, 51)
            self.assertEqual(downsampled_data.pop(0).size, 50)
            self.assertEqual(downsampled_data.pop(-1).size, 51)

        with self.subTest('test_intermediate_data_points_not_removed'):
            self.assertTrue(
                all(times.size == 100 for times in downsampled_times)
            )
            self.assertTrue(
                all(data.size == 100 for data in downsampled_times)
            )

    def test_trace_list_empty(self):
        self.trimmed_traces_list = []
        downsampled_times, downsampled_data = downsample_waveform_data(
            self.trimmed_traces_list,
            self.start_time, self.end_time
        )
        self.assertListEqual(downsampled_times, [])
        self.assertListEqual(downsampled_data, [])

    def test_end_time_earlier_than_start_time(self):
        self.start_time, self.end_time = self.end_time, self.start_time
        downsampled_times, downsampled_data = downsample_waveform_data(
            self.trimmed_traces_list,
            self.start_time, self.end_time
        )
        self.assertTrue(all(times.size == 0 for times in downsampled_times))
        self.assertTrue(all(data.size == 0 for data in downsampled_data))

    @patch('sohstationviewer.model.handling_data.downsample')
    def test_arguments_sent_to_downsample(self, mock_downsample):
        mock_downsample.return_value = (1, 2, 3)
        const.CHAN_SIZE_LIMIT = 1000
        downsample_waveform_data(self.trimmed_traces_list,
                                 self.start_time, self.end_time)

        positional_args, named_args = mock_downsample.call_args
        self.assertEqual(len(positional_args), 2)


class TestDownsample(TestCase):
    def setUp(self) -> None:
        patcher = patch('sohstationviewer.model.downsampler.chunk_minmax')
        self.addCleanup(patcher.stop)
        self.mock_chunk_minmax = patcher.start()
        self.times = np.arange(1000)
        self.data = np.arange(1000)
        self.log_idx = np.arange(1000)

    def test_first_downsample_step_remove_enough_points(self):
        req_points = 999
        downsample(self.times, self.data, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)

    def test_first_downsample_step_remove_enough_points_with_logidx(self):
        req_points = 999
        downsample(self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)

    def test_second_downsample_step_required(self):
        req_points = 1
        downsample(self.times, self.data, rq_points=req_points)
        self.assertTrue(self.mock_chunk_minmax.called)
        times, data, _, rq_points = self.mock_chunk_minmax.call_args[0]
        self.assertIsNot(times, self.times)
        self.assertIsNot(data, self.data)
        self.assertEqual(rq_points, req_points)

    def test_second_downsample_step_required_with_logidx(self):
        req_points = 1
        downsample(self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertTrue(self.mock_chunk_minmax.called)
        times, data, log_idx, rq_points = self.mock_chunk_minmax.call_args[0]
        self.assertIsNot(times, self.times)
        self.assertIsNot(data, self.data)
        self.assertIsNot(log_idx, self.log_idx)
        self.assertEqual(rq_points, req_points)

    def test_requested_points_greater_than_data_size(self):
        req_points = 10000
        times, data, _ = downsample(
            self.times, self.data, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)
        # Check that we did not do any processing on the times and data arrays.
        # This ensures that we don't do two unneeded copy operations.
        self.assertIs(times, self.times)
        self.assertIs(data, self.data)

    def test_requested_points_greater_than_data_size_with_logidx(self):
        req_points = 10000
        times, data, log_idx = downsample(
            self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)
        # Check that we did not do any processing on the times and data arrays.
        # This ensures that we don't do two unneeded copy operations.
        self.assertIs(times, self.times)
        self.assertIs(data, self.data)
        self.assertIs(log_idx, self.log_idx)

    def test_requested_points_is_zero(self):
        req_points = 0
        downsample(self.times, self.data, rq_points=req_points)
        self.assertTrue(self.mock_chunk_minmax.called)
        times, data, _, rq_points = self.mock_chunk_minmax.call_args[0]
        self.assertIsNot(times, self.times)
        self.assertIsNot(data, self.data)
        self.assertEqual(rq_points, req_points)

    def test_requested_points_is_zero_with_logidx(self):
        req_points = 0
        downsample(self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertTrue(self.mock_chunk_minmax.called)
        times, data, log_idx, rq_points = self.mock_chunk_minmax.call_args[0]
        self.assertIsNot(times, self.times)
        self.assertIsNot(data, self.data)
        self.assertIsNot(log_idx, self.log_idx)
        self.assertEqual(rq_points, req_points)

    def test_empty_times_and_data(self):
        req_points = 1000
        self.times = np.empty((0, 0))
        self.data = np.empty((0, 0))
        times, data, _ = downsample(
            self.times, self.data, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)
        # Check that we did not do any processing on the times and data arrays.
        # This ensures that we don't do two unneeded copy operations.
        self.assertIs(times, self.times)
        self.assertIs(data, self.data)

    def test_empty_times_and_data_with_logidx(self):
        req_points = 1000
        self.times = np.empty((0, 0))
        self.data = np.empty((0, 0))
        self.log_idx = np.empty((0, 0))
        times, data, log_idx = downsample(
            self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)
        # Check that we did not do any processing on the times and data arrays.
        # This ensures that we don't do two unneeded copy operations.
        self.assertIs(times, self.times)
        self.assertIs(data, self.data)
        self.assertIs(log_idx, self.log_idx)


class TestChunkMinmax(TestCase):
    def setUp(self):
        self.times = np.arange(1000)
        self.data = np.arange(1000)
        self.log_idx = np.arange(1000)

    def test_data_size_is_multiple_of_requested_points(self):
        req_points = 100
        times, data, log_idx = chunk_minmax(
            self.times, self.data, self.log_idx, req_points)
        self.assertEqual(times.size, req_points)
        self.assertEqual(data.size, req_points)
        self.assertEqual(log_idx.size, req_points)

    @patch('sohstationviewer.model.downsampler.downsample', wraps=downsample)
    def test_data_size_is_not_multiple_of_requested_points(
            self, mock_downsample):
        req_points = 102
        chunk_minmax(self.times, self.data, self.log_idx, req_points)
        self.assertTrue(mock_downsample.called)

    def test_requested_points_too_small(self):
        small_req_points_list = [0, 1]
        for req_points in small_req_points_list:
            with self.subTest(f'test_requested_points_is_{req_points}'):
                times, data, log_idx = chunk_minmax(
                    self.times, self.data, self.log_idx, rq_points=req_points)
                self.assertEqual(times.size, 0)
                self.assertEqual(data.size, 0)
                self.assertEqual(data.size, 0)


class TestTrimDownsampleSohChan(TestCase):
    @staticmethod
    def downsample(times, data, log_indexes=None, rq_points=0):
        return times, data, log_indexes

    def setUp(self) -> None:
        self.channel_info = {}
        self.traces_info = [{
            'times': np.arange(1000),
            'data': np.arange(1000)
        }]
        self.channel_info['tracesInfo'] = self.traces_info
        self.start_time = 250
        self.end_time = 750
        self.first_time = False

        patcher = patch('sohstationviewer.model.handling_data.downsample')
        self.addCleanup(patcher.stop)
        self.mock_downsample = patcher.start()
        self.mock_downsample.side_effect = self.downsample

    def num_points_outside_time_range(self, start_time, end_time):
        return len([data_point
                    for data_point in self.traces_info[0]['times']
                    if not start_time <= data_point <= end_time])

    def test_start_time_later_than_times_data(self):
        self.start_time = 250
        self.end_time = 1250
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        self.assertGreaterEqual(self.channel_info['times'][0].min(),
                                self.start_time)
        self.assertEqual(
            self.traces_info[0]['times'].size -
            self.channel_info['times'][0].size,
            self.num_points_outside_time_range(self.start_time, self.end_time)
        )

    def test_end_time_earlier_than_times_data(self):
        self.start_time = -250
        self.end_time = 750
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        self.assertLessEqual(self.channel_info['times'][0].max(),
                             self.end_time)
        self.assertEqual(
            self.traces_info[0]['times'].size -
            self.channel_info['times'][0].size,
            self.num_points_outside_time_range(self.start_time, self.end_time)
        )

    def test_start_time_earlier_than_times_data(self):
        self.start_time = -250
        self.end_time = 750
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        self.assertEqual(self.channel_info['times'][0].min(),
                         self.channel_info['tracesInfo'][0]['times'].min())
        self.assertEqual(
            self.traces_info[0]['times'].size
            - self.channel_info['times'][0].size,
            self.num_points_outside_time_range(self.start_time, self.end_time)
        )

    def test_end_time_later_than_times_data(self):
        self.start_time = 250
        self.end_time = 1250
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        self.assertEqual(self.channel_info['times'][0].max(),
                         self.channel_info['tracesInfo'][0]['times'].max())
        self.assertEqual(
            self.traces_info[0]['times'].size -
            self.channel_info['times'][0].size,
            self.num_points_outside_time_range(self.start_time, self.end_time)
        )

    def test_times_data_contained_in_time_range(self):
        self.start_time = -250
        self.end_time = 1250
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        np.testing.assert_array_equal(self.channel_info['times'][0],
                                      self.traces_info[0]['times'])

    def test_time_range_is_the_same_as_times_data(self):
        self.start_time = ZERO_EPOCH_TIME
        self.end_time = 999
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        np.testing.assert_array_equal(self.channel_info['times'][0],
                                      self.traces_info[0]['times'])

    def test_time_range_does_not_overlap_times_data(self):
        self.start_time = 2000
        self.end_time = 3000
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        self.assertEqual(len(self.channel_info['times']), 0)
        self.assertEqual(len(self.channel_info['data']), 0)

    def test_data_is_downsampled(self):
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        self.assertTrue(self.mock_downsample.called)

    def test_processed_data_is_stored_in_appropriate_location(self):
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        expected_keys = ('tracesInfo', 'times', 'data')
        self.assertTupleEqual(tuple(self.channel_info.keys()),
                              expected_keys)

    @patch('sohstationviewer.model.handling_data.downsample')
    def test_arguments_sent_to_downsample(self, mock_downsample):
        mock_downsample.return_value = (1, 2, 3)
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)

        positional_args, named_args = mock_downsample.call_args
        self.assertEqual(len(positional_args), 2)


class TestTrimDownsampleSohChanWithLogidx(TestCase):
    def setUp(self) -> None:
        self.channel_info = {'reftek': True}
        self.traces_info = [{
            'times': np.arange(1000),
            'data': np.arange(1000),
            'logIdx': np.arange(1000)
        }]
        self.channel_info['tracesInfo'] = self.traces_info
        self.start_time = 250
        self.end_time = 750
        self.first_time = False

    def test_time_range_does_not_overlap_times_data(self):
        self.start_time = 2000
        self.end_time = 3000
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        self.assertEqual(len(self.channel_info['times']), 0)
        self.assertEqual(len(self.channel_info['data']), 0)
        self.assertEqual(len(self.channel_info['logIdx']), 0)

    def test_processed_data_is_stored_in_appropriate_location(self):
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)
        expected_keys = ('reftek', 'tracesInfo', 'times', 'data', 'logIdx')
        self.assertTupleEqual(tuple(self.channel_info.keys()),
                              expected_keys)

    @patch('sohstationviewer.model.handling_data.downsample')
    def test_arguments_sent_to_downsample(self, mock_downsample):
        mock_downsample.return_value = (1, 2, 3)
        trim_downsample_chan_with_spr_less_or_equal_1(
            self.channel_info, self.start_time, self.end_time)

        positional_args, named_args = mock_downsample.call_args
        self.assertEqual(len(positional_args), 3)


class TestTrimDownsampleWfChan(TestCase):
    def no_file_memmap(self, file_path: Path, **kwargs):
        # Data will look the same as times. This has two benefits:
        # - It is a lot easier to inspect what data remains after trimming
        # and downsampling, seeing as the remaining data would be the same
        # as the remaining times.
        # - It is a lot easier to reproducibly create a test data set.
        array_size = 100
        file_idx = int(file_path.name.split('-')[-1])
        start = file_idx * array_size
        end = start + array_size
        return np.arange(start, end)

    def setUp(self) -> None:
        memmap_patcher = patch.object(np, 'memmap',
                                      side_effect=self.no_file_memmap)
        self.addCleanup(memmap_patcher.stop)
        memmap_patcher.start()

        self.channel_data = {}
        self.traces_info = []
        self.channel_data['tracesInfo'] = self.traces_info
        self.data_folder = TemporaryDirectory()
        for i in range(100):
            trace_size = 100
            start_time = i * trace_size
            trace = {}
            trace['startTmEpoch'] = start_time
            trace['endTmEpoch'] = start_time + trace_size - 1
            trace['size'] = trace_size

            times_file_name = Path(self.data_folder.name) / f'times-{i}'
            trace['times_f'] = times_file_name

            data_file_name = Path(self.data_folder.name) / f'data-{i}'
            trace['data_f'] = data_file_name

            self.traces_info.append(trace)
        self.start_time = 2500
        self.end_time = 7500

    def test_result_is_stored(self):
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, True)
        self.assertTrue('times' in self.channel_data)
        self.assertGreater(len(self.channel_data['times']), 0)
        self.assertTrue('data' in self.channel_data)
        self.assertGreater(len(self.channel_data['data']), 0)

    def test_data_small_enough_after_first_trim_flag_is_set(self):
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, True)
        self.assertTrue('fulldata' in self.channel_data)

    def test_no_additional_work_if_data_small_enough_after_first_trim(self):
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, True)
        current_times = self.channel_data['times']
        current_data = self.channel_data['data']
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, True)
        self.assertIs(current_times, self.channel_data['times'])
        self.assertIs(current_data, self.channel_data['data'])

    def test_data_too_large_after_trimming(self):
        const.RECAL_SIZE_LIMIT = 1
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, False)
        self.assertTrue('times' not in self.channel_data)
        self.assertTrue('data' not in self.channel_data)
        const.RECAL_SIZE_LIMIT = ORIGINAL_RECAL_SIZE_LIMIT

    @patch('sohstationviewer.model.handling_data.trim_waveform_data',
           wraps=trim_waveform_data)
    @patch('sohstationviewer.model.handling_data.downsample_waveform_data',
           wraps=downsample_waveform_data)
    def test_data_trim_and_downsampled(self, mock_downsample, mock_trim):
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, False)
        self.assertTrue(mock_trim.called)
        self.assertTrue(mock_downsample.called)
